"""
Requirements:
Add a way to set a default system to be accessed. You must be able to override this by explicitly specifying system id in command
add intuitive file access and manipulation using this default system. 
"""


import json
import os
import webbrowser


if __name__ != "__main__":
    from . import baseCommand, decorators
    from .arguments import argument
    from . import commandOpts
    from socketopts import schemas
    Argument = argument.Argument


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
base_config_path = os.path.join(__location__, '..\\..\\..\\tapis-config-files\\system-config.json')


class system(baseCommand.BaseCommand):
    """
    @help: set the system to run operations on by default. Running this will put you "in" the system so that you dont have to specify system ID for each command
    """
    required_arguments=[
        Argument('systemId', size_limit=(1, 80)),
    ]
    async def run(self, *args, **kwargs):
        system_info = self.t.systems.getSystem(systemId=kwargs['systemId'])
        kwargs['connection'].system = kwargs['systemId']
        kwargs['connection'].pwd = system_info.rootDir
        return f"successfully entered the system {kwargs['connection'].system}"


class exit_system(baseCommand.BaseCommand):
    """
    @help: exit the default system
    """
    async def run(self, *args, **kwargs):
        kwargs['connection'].system = ''
        kwargs['connection'].pwd = ''
        return "successfully exited system"
    

class get_systems(baseCommand.BaseCommand):
    """
    @help: Gets and returns the list of systems the current Tapis service and account have access to
    @doc: this is an example of the doc segment of the docstring. not included in help message
    """
    async def run(self, *args, **kwargs):
        systems = self.t.systems.getSystems()
        return systems
    

class get_system_info(baseCommand.BaseCommand):
    """
    @help: get information on a selected system
    """
    required_arguments=[
        Argument('systemId', size_limit=(1, 80))
    ]
    async def run(self, *args, **kwargs): 
        system_info = self.t.systems.getSystem(systemId=kwargs['systemId'])
        return system_info
    

class verify_pki_keys(baseCommand.BaseCommand):
    """
    @help: upload system credentials to a system. Must generate keys first using 'ssh-keygen -m PEM -f id_rsa', and format with, 'awk -v ORS='\\n' '1' <private_key_name>
    file argument must contain the path to the private and public keys respectively, separated by a ','
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('systemId'),
        Argument('private_key_path'),
        Argument('public_key_path')
    ]
    async def run(self, *args, **kwargs): # get information about a system given its ID
        with open(kwargs['private_key_path'], 'r') as f:
            private_key = f.read()

        with open(kwargs['public_key_path'], 'r') as f:
            public_key = f.read()

        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['systemId'],
                            userName=self.username,
                            privateKey=private_key,
                            publicKey=public_key)

        return str(cred_return_value)
    

class get_scheduler_profiles(baseCommand.BaseCommand):
    """
    @help: get list of scheduler profiles in order to use while creating system
    """
    async def run(self, *args, **kwargs):
        return [{"name":scheduler.name, "description":scheduler.description, "tenant":scheduler.tenant} for scheduler in self.t.systems.getSchedulerProfiles()]


class get_scheduler_profiles_choices(argument.DynamicChoiceList):
    def __call__(self, tapis_instance):
        profiles_unfiltered = tapis_instance.systems.getSchedulerProfiles()
        return [profile.name for profile in profiles_unfiltered]
    

class create_system(baseCommand.BaseCommand):
    """
    @help: create a system. Must have a properly configured system file.
    see the template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/system-config.json
    this command will automatically create and upload the ssh keys
    """
    supports_config_file=True
    required_arguments=[
        Argument('systemId', size_limit=(1, 80)),
        Argument('systemType', choices=["LINUX", "S3", "IRODS", "GLOBUS"], description=
                                    """LINUX is a standard linux kernel
                                    S3 refers to an AWS S3 Bucket
                                    IRODS refers to an IRODS data management system
                                    GLOBUS refers to a GLOBUS file system"""),
        Argument('host', size_limit=(1, 256), description="In the case of Linux this is the hostname or IP of the HPC system you want to connect to. For S3, this is the AWS bucket URL"),
        Argument('defaultAuthnMethod', choices=['PASSWORD', "PKI_KEYS", "ACCESS_KEY", "TOKEN", "CERT"], description=
                                    """Depending on your systemType, you will be restricted to certain options.
                                    Linux: PASSWORD, PKI_KEYS
                                    S3: ACCESS_KEY
                                    GLOBUS: TOKEN
                                    IRODS: TOKEN
                                    In the case you choose password, your username and password will either be your TACC account info, or the login info you used with federated/device_code grant"""),
        Argument('canExec', action='store_true'),
    ]
    optional_arguments=[
        Argument('description', arg_type='str_input', size_limit=(0, 2048)),
        Argument('owner'),
        Argument('enabled', action='store_true'),
        Argument('effectiveUserId', default_value=r"${apiUserId}", size_limit=(0, 60)),
        Argument('bucketName'),
        Argument('rootDir', default_value='/', size_limit=(0, 4096)),
        Argument('port', data_type='int'),
        Argument('useProxy', action='store_true'),
        Argument('proxyHost', size_limit=(0, 256)),
        Argument('proxyPort', data_type='int'),
        Argument('isDtn', action='store_true'),
        Argument('dtnSystemId', size_limit=(0, 80)),
        Argument('dtnMountPoint'),
        Argument('canRunBatch', action='store_true'),
        Argument('enableCmdPrefix', action='store_true'),
        Argument('mpiCmd', size_limit=(0, 126), arg_type='str_input'),
        Argument('jobRuntimes', arg_type='input_list', data_type=argument.Form(
            'jobRuntime', arguments_list = [
                Argument('runtimeType', choices=['DOCKER', 'SINGULARITY']), 
                Argument('version')
                ]
            )),
        Argument('jobWorkingDir', default_value=r"HOST_EVAL($WORK2)", size_limit=(0, 4096)),
        Argument('jobEnvVariables', arg_type='input_list', data_type=argument.Form(
            'jobEnvironmentVariable', arguments_list = [
                Argument('key'),
                Argument('value', default_value=''),
                Argument('description', size_limit=(0, 2048), arg_type='str_input')
            ]
        )),
        Argument('jobMaxJobs', data_type='int'),
        Argument('jobMaxJobsPerUser', data_type='int'),
        Argument('batchScheduler', choices=['SLURM', "CONDOR", "PBS", "SGE", "UGE", "TORQUE"]),
        Argument('batchLogicalQueues', arg_type='input_list', data_type=argument.Form(
            'batchLogicalQueue', arguments_list = [
                Argument('name', size_limit=(1, 128)),
                Argument('hpcQueueName', size_limit=(1, 128)),
                Argument('maxJobs', data_type='int'),
                Argument('maxJobsPerUser', data_type='int'),
                Argument('minNodeCount', data_type='int'),
                Argument('maxNodeCount', data_type='int'),
                Argument('minCoresPerNode', data_type='int'),
                Argument('maxCoresPerNode', data_type='int'),
                Argument('minMemoryMB', data_type='int'),
                Argument('maxMemoryMB', data_type='int'),
                Argument('minMinutes', data_type='int'),
                Argument('maxMinutes', data_type='int')
            ]
        )),
        Argument('batchDefaultLoginQueue', size_limit=(1, 128)),
        Argument('batchSchedulerProfile', choices=get_scheduler_profiles_choices()),
        Argument('jobCapabilities', arg_type='input_list', data_type=argument.Form(
            'jobCapability', arguments_list=[
                Argument('category', choices=['SCHEUDLER', 
                                              'OS', 'HARDWARE', 
                                              'SOFTWARE', 'JOB', 
                                              'CONTAINER', 'MISC', 
                                              'CUSTOM']),
                Argument('name', size_limit=(1, 128)),
                Argument('datatype', choices=['STRING', 'INTEGER', 
                                              'BOOLEAN', 'NUMBER', 
                                              'TIMESTAMP']),
                Argument('precedence', data_type='int'),
                Argument('value')
            ]
        )),
        Argument('tags', arg_type='input_list'),
        Argument('notes', arg_type='str_input'),
        Argument('importRefId')
    ]
    def __init__(self):
        super().__init__()
        self.sys_auth_map = {"LINUX":{"PASSWORD":self.password_auth, "PKI_KEYS":self.pki_keys_auth}, 
                             "S3":{"ACCESS_KEY":self.access_key_auth}, "GLOBUS":{"TOKEN":self.token_auth}, 
                             "IRODS":{"TOKEN":self.token_auth}}

    async def password_auth(self, **kwargs):
        userName = self.username
        if self.server.auth_type == 'password':
            password = self.password
        else:
            request = schemas.FormRequest(request_content={"password":Argument('password', arg_type='secure')},
                                          message=f"Please enter the credentials you entered on the web portal {self.server.url}")
            await kwargs['connection'].send(request)
            response = await kwargs['connection'].receive()
            response_content = response.request_content
            password = response_content['password']
        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['id'],
                            userName=userName,
                            password=password)
        return cred_return_value
    
    async def token_auth(self, **kwargs):
        session_details = self.systems.getGlobusAuthUrl()
        request = schemas.FormRequest(message=f"travel to the url {session_details.url} if the page doesnt open on its own, and enter your credentials. Then paste the code displayed to this app",
                                      request_content={'Auth_code':Argument('code', arg_type='str_input')})
        webbrowser.open(session_details.url)
        await kwargs['connection'].send(request)
        response = await kwargs['connection'].receive()
        auth_code = response.request_content['Auth_code']
        token_info = self.t.systems.generateGlobusTokens(systemId=kwargs['id'], userName=self.username, authCode=auth_code, sessionId=session_details.sessionId)
        return str(token_info)
    
    async def access_key_auth(self, **kwargs):
        request = schemas.FormRequest(message=f"Retrieve the access key and access secret from your S3 bucket",
                                      request_content={'access_key':Argument('access_key', arg_type='str_input'), 'access_secret':Argument('access_secret', arg_type='secure')})
        await kwargs['connection'].send(request)
        response = await kwargs['connection'].receive()
        cred_return_value = self.t.systems.createUserCredential(systemId=kwargs['id'],
                            accessKey=response.request_content['access_key'],
                            password=response.request_content['access_secret'])
        return cred_return_value
    
    async def pki_keys_auth(self, **kwargs):
        request = schemas.FormRequest(message=
                                      f"""Follow these steps to manually register PKI keys with the system
                                      1. ssh into the host you are creating your system on, {kwargs['host']}
                                      2. navigate into the .ssh folder of the host. If there are not pre-generated keys, run ssh-keygen -t rsa -b 4096 -m PEM
                                      3. format both key files using cat $privateKeyFile | awk -v ORS='\\n' '1'. Both keys must be in single line format
                                      4. ensure the private key is not in openSSH format (that is it has the begin and end headers)
                                      5. download both to your local machine using scp
                                      6. run the verify_pki_keys command and submit the system id {kwargs['id']} and the file path of the public and private keys
                                      7. you should be able to access the system now""",
                                      request_content={"continue":Argument("continue", arg_type='confirmation')})
        await kwargs['connection'].send(request)
        await kwargs['connection'].receive()
        return None        

    async def run(self, *args, **kwargs) -> str: # create a tapius system. Takes a path to a json file with all system information, as well as an ID
        system_creation_info = self.t.systems.createSystem(**kwargs)
        if kwargs['defaultAuthnMethod'] not in self.sys_auth_map[kwargs['systemType']]:
            raise ValueError(f"The system type {kwargs['systemType']} does not support {kwargs['defaultAuthnMethod']} authentication.")
        else:
            auth_result = self.sys_auth_map[kwargs['systemType']]['defaultAuthnMethod'](**kwargs)
        return system_creation_info
    

class update_system(create_system):
    """
    @help: update a system with new information
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    async def run(self, *args, **kwargs):
        result = self.t.putSystem(**kwargs)
        return result
    

class is_enabled(baseCommand.BaseCommand):
    """
    @help: check to see if a system is enabled
    """
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments = [
        Argument('systemId')
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.isEnabled(**kwargs)


class enable_system(is_enabled):
    """
    @help: enable a system
    """
    async def run(self, *args, **kwargs):
        return self.t.systems.enableSystem(**kwargs)
    

class disable_system(is_enabled):
    """
    @help: disable a system
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, *args, **kwargs):
        return self.t.systems.disableSystem(**kwargs)
    

class delete_system(baseCommand.BaseCommand):
    """
    @help: delete the selected system
    """
    decorator=decorators.NeedsConfirmation()
    command_opt = [commandOpts.CHECK_EXPLICIT_ID('systemId')]
    required_arguments=[
        Argument('systemId', size_limit=(1, 80)),
    ]
    async def run(self, *args, **kwargs) -> str:
        kwargs['connection'].system = ''
        kwargs['connection'].pwd = ''
        return_value = self.t.systems.deleteSystem(systemId=kwargs['systemId'])
        return return_value
    

class undelete_system(baseCommand.BaseCommand):
    """
    @help: undo deletion
    """
    required_arguments=[
        Argument('systemId', size_limit=(1, 80)),
    ]
    async def run(self, *args, **kwargs):
        return self.t.systems.undelete(**kwargs)

class create_child_system(baseCommand.BaseCommand):
    """
    @help: create a child system which inherits majority attributes from parent
    """
    async def run(self, *args, **kwargs):
        return "UNFINISHED FEATURE"


from getpass import getpass
import os
import json

from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit import prompt
from blessed import Terminal

if __name__ != "__main__":
    from ..socketopts import schemas


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
saved_command = os.path.join(__location__, r'entered_command.json')


class Formatters:
    """
    Format received dictionaries in the client code
    """
    def print_response(self, input_data, depth: int=0):
        if isinstance(input_data, (list, set, tuple)):
            for data in input_data:
                if isinstance(data, (list, dict, set, tuple)):
                    print("\n")
                self.print_response(data, depth=depth+1)
        elif isinstance(input_data, dict):
            for name, data in input_data.items():
                if isinstance(data, (int, str)):
                    print(("  " * depth) + f"{name}: {data}")
                    continue
                print(("  " * depth) + f"{name}: ")
                self.print_response(data, depth=depth+1)
        elif isinstance(input_data, (int, str)):
            print(input_data)
        if depth == 0:
            print("\n")


def cast_int(data):
    return int(data)

def cast_string(data):
    return str(data)


class ParserTypeLenEnforcer:
        type_map = {'int':cast_int, 'string':cast_string}
        def __init__(self, name: str=str(), size: tuple=(0, 0), data_type: str='string', choices: list=list()):
            self.arg_name = name
            self.update_data_type(data_type)
            self.lower_size_limit, self.upper_size_limit = size
            self.choices = choices

        def update_data_type(self, d_type_str):
            try:
                self.data_type = self.type_map[d_type_str]
            except KeyError:
                self.data_type = None

        def update_constraints(self, name=None, data_type=None, choices=None, size_limit=None, **kwargs):
            self.arg_name = name
            self.data_type = self.update_data_type(data_type)
            self.choices = choices
            self.lower_size_limit, self.upper_size_limit = size_limit

        def __call__(self, data):
            if self.data_type:
                try:
                    self.data_type(data)
                except Exception as e:
                    raise ValidationError(message=str(e) + self.data_type, cursor_position=0)
            if self.data_type == int:
                if not data >= self.lower_size_limit or not data < self.upper_size_limit:
                    raise ValidationError(message=f"The input for the argument {self.arg_name} must be in the range ({self.lower_size_limit}, {self.upper_size_limit}). Got value {data}", cursor_position=0)
            elif self.data_type == str:
                if not len(data) >= self.lower_size_limit or not len(data) < self.upper_size_limit:
                    raise ValidationError(message=f"The input length for the argument {self.arg_name} must be in the range ({self.lower_size_limit}, {self.upper_size_limit}). Got length {len(data)}", cursor_position=0)
            if self.choices and data not in self.choices:
                raise ValidationError(message=f"The input for the argument {self.arg_name} must be one of the following: {self.choices}", cursor_position=0)


class ResponseValidator(Validator):
    def __init__(self):
        super().__init__()
        self.enforcer = ParserTypeLenEnforcer()

    def validate(self, document):
        text = document.text
        self.enforcer(text)
            

class Handlers(Formatters):
    def __init__(self):
        self.validator = ResponseValidator()

    def __expression_input(self) -> str: 
        """
        Input an expression as requested by the server for something like cypher queries
        """
        print("Enter 'exit' to submit") 
        expression = ''
        line = ''
        while line != 'exit': 
            line = str(input("> "))
            if line != 'exit':
                expression += line
        return expression
    
    def confirmation_handler(self):
        print("are you sure?")
        while True:
            decision = str(input("(y/n)"))
            if decision == 'y':
                decision = True
                break
            elif decision == 'n':
                decision = False
                break
            else:
                print("Enter valid response")
        return decision
    
    def form_handler(self, form_request: dict, term: Terminal):
        response = dict()
        repeat = False
        for field, attrs in form_request.items():
            while True:
                arg_type = attrs['arg_type']
                self.validator.enforcer.update_constraints(**attrs)
                try:
                    match arg_type:
                        case 'secure':
                            answer = prompt(f"{attrs['name']}: ", validator=self.validator, is_password=True)
                        case 'expression':
                            with term.fullscreen():
                                print(f"Enter expression input for the {attrs['name']} argument.")
                                answer = self.__expression_input()
                        case 'form':
                            answer = self.form_handler(attrs['arguments_list'], term)
                        case 'input_list':
                            repeat = True
                            answer = []
                            with term.fullscreen():
                                print(f"You are now entering data for the {attrs['args']} field")
                                while repeat:
                                    sub_answer, repeat = self.form_handler(attrs['data_type'], term)
                                    answer.append(sub_answer)
                        case 'input_dict':
                            repeat = True
                            answer = dict()
                            with term.fullscreen():
                                print(f"You are now entering data for the {attrs['args']} field")
                                while repeat:
                                    name = str(input(f"enter the {attrs['data_type']['name']} for the instance of {attrs['name']}"))
                                    sub_answer, repeat = self.form_handler(attrs['data_type'], term)
                                    answer[name] = sub_answer
                        case 'str_input':
                            answer = prompt(f"{attrs['name']}: ", validator=self.validator)
                        case 'confirmation':
                            answer = self.confirmation_handler()
                        case 'silent':
                            continue
                        case _:
                            raise AttributeError(f"There is not argument type {arg_type}")
                    response[field] = answer
                    break
                except KeyboardInterrupt:
                    raise RuntimeError('Form cancelled, command execution stopped')
                except Exception as e:
                    with open(saved_command, 'w') as f:
                        f.write(json.dumps(response))
                        print(f"Argument input failure, command data written to file {saved_command}")
                        raise e
        return response, repeat
    
    def universal_message_handler(self, message: schemas.BaseSchema, term: Terminal):
        self.print_response(message.message)
        if message.error:
            self.print_response(message.error)
        if message.request_content:
            filled_form, repeat = self.form_handler(message.request_content, term)
            return filled_form, repeat
        return None, None
        
    def environment_cli_response_stream_handler(self, response):
        if response.schema_type == 'ResponseData' and response.exit_status: # if the command was a shutdown or exit, close the program
            self.print_response(response.response_message)
            os._exit(0)
        elif response.schema_type == 'ResponseData':
            if response.active_username:
                self.username = response.active_username
            if response.url:
                self.url = response.url
            self.print_response(response.response_message)


import typing

from tapipy.tapis import Tapis


if __name__ != "__main__":
    from . import baseCommand, decorators
    from utilities import exceptions
    from commands.arguments.argument import Argument


class switch_service(baseCommand.BaseCommand):
    """
    @help: switch the connected tapis service
    @todo: upgrade to federated auth
    """
    required_arguments=[
        Argument('link', size_limit=(0, 80)),
        Argument('auth', choices=['password', 'device_code', 'federated']),
        Argument('connection', arg_type='silent')
    ]
    async def run(self, *args, **kwargs):  # link is the baseURL
        auth = kwargs['auth']
        link = kwargs['link']
        self.server.auth_type = auth
        if auth == "password":
            results = await self.server.password_grant(link, kwargs['connection'])
        elif auth == "device_code":
            results = await self.server.device_code_grant(link, kwargs['connection'])
        elif auth == "federated":
            results = await self.server.federated_grant(link, kwargs['connection'])
        else:
            results = None
        self.server.configure_decorators(self.server.username, self.server.password)
        self.server.update_credentials(self.server.t, self.server.username, self.server.password)
        return results
      

class exit(baseCommand.BaseCommand):
    """
    @help: exit the CLI without shutting down the service
    """
    async def run(self, *args, **kwargs):
        raise exceptions.Exit
    

class shutdown(baseCommand.BaseCommand):
    """
    @help: exit the CLI and shutdown the service
    """
    decorator = decorators.NeedsConfirmation()
    async def run(self, *args, **kwargs):
        raise exceptions.Shutdown
    
    
class whoami(baseCommand.BaseCommand):
    """
    @help: returns the username of the current user
    """
    async def run(self, *args, **kwargs) -> str:
        user_info = self.t.authenticator.get_userinfo()
        return user_info
    

class get_args(baseCommand.BaseCommand):
    """
    @help: get the list of possible arguments 
    """
    async def run(self, *args, **kwargs):
        return self.server.arguments
    

class whereami(baseCommand.BaseCommand):
    """
    @help: get the URI of current tapis tenant
    """
    async def run(self, *args, **kwargs):
        return self.server.url


import socket
import asyncio
from TapisCLICICLE.client.cli import CLI

def client_start():
    client_object = CLI(socket.gethostbyname(socket.gethostname()), 30000)
    client_object.main()

client_start()

# server = Server(socket.gethostbyname(socket.gethostname()), 30000)
# asyncio.run(server.main())


#!/usr/bin/python3
import socket
import argparse
import sys
import os
import time
import traceback

import pyfiglet
from blessed import Terminal

if __name__ != "__main__":
    from . import handlers
    from ..socketopts import socketOpts, schemas
    from ..commands import decorators
    from ..utilities import killableThread


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(f"{__file__}")))
server_path = os.path.join(__location__, r'..\serverRun.py')
    

class ClientSideConnection(socketOpts.ClientSocketOpts, handlers.Handlers):
    def __init__(self, connection, debug=False):
        super().__init__("client", debug=debug)
        self.connection = connection
    
    def close(self):
        self.connection.close()


class CLI(handlers.Handlers):
    debug=False
    """
    Receive user input, either direct from bash environment or from the custom interface, then parse these commands and send them to the server to be executed. 
    """
    def __init__(self, IP: str, PORT: int):
        super().__init__()

        self.term = Terminal()
        self.ip, self.port = IP, PORT
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 

        # set up argparse
        self.parser = None

        setup_message = schemas.ResponseData(request_content={'setup_success':True})
        try:
            self.username, self.url = self.connect()
            self.connection.send(setup_message)
        except (KeyboardInterrupt, Exception) as e:
            error_str = traceback.format_exc()
            if self.debug:
                print(error_str)
            print(e)
            setup_message.error = str(e)
            setup_message.request_content['setup_success'] = False
            self.connection.send(setup_message)
            os._exit(0)

        self.pwd = ""
        self.current_system = ""

    def parser_error(self, args):
        print(f"Ignoring unrecognized arguments: {args}")
    
    def configure_parser(self, arguments):
        parser = argparse.ArgumentParser(description="Command Line Argument Parser", exit_on_error=False, usage=argparse.SUPPRESS, add_help=False, conflict_handler='resolve')
        parser.add_argument('command_selection')
        parser.add_argument('positionals', nargs='*', default='default1')
        parser.error = self.parser_error

        for arg_name, arg in arguments.items():
            print(arg['truncated_arg'])
            parser.add_argument(f"-{arg['truncated_arg']}", arg['full_arg'],
                                action=arg['action'])
        return parser

    def initialize_server(self): 
        """
        detect client operating system. The local server intitialization is different between unix and windows based systems
        """
        if 'win' in sys.platform:
            os.system(rf"pythonw {server_path}")
        else: # unix based
            os.system(f"python {server_path} &")

    #@decorators.AnimatedLoading
    def connection_initialization(self):
        """
        start the local server through the client
        """
        startup_flag = False
        timeout_time = time.time() + 30 # server setup timeout. If expires, there is a problem!
        while True:
            if time.time() > timeout_time: # connection timeout condition
                sys.stdout.write("\r[-] Connection timeout")
                os._exit(0)
            try:
                self.connection.connect((self.ip, self.port)) 
                self.connection = ClientSideConnection(self.connection, debug=self.debug)
                if startup_flag:
                    startup.kill()
                break
            except Exception as e:
                if not startup_flag:
                    startup = killableThread.KillableThread(target=self.initialize_server) # run the server setup on a separate thread
                    startup.start() 
                    startup_flag = True # set the flag to true so the thread runs only once
                    continue

    def auth(self):
        while True:
            server_auth_request: schemas.AuthRequest = self.connection.receive()
            if not server_auth_request.message and not server_auth_request.request_content:
                self.print_response(server_auth_request.error)
                sys.exit(0)
            elif server_auth_request.auth_request_type == "success":
                self.print_response(server_auth_request.message)
                return server_auth_request.message['username'], server_auth_request.message['url']
            form_response, repeat = self.universal_message_handler(server_auth_request, self.term)
            if not form_response:
                break
            client_response_request = schemas.AuthRequest(auth_request_type=server_auth_request.auth_request_type, request_content=form_response)
            self.connection.send(client_response_request)
        
    def connect(self):
        """
        connect to the local server
        """
        self.connection_initialization() 
        #self.connection.connect((self.ip, self.port)) # enable me for debugging. Requires manual server start
        connection_info: schemas.StartupData = self.connection.receive() # receive info from the server whether it is a first time connection
        if connection_info.initial: # if the server is receiving its first connection for the session\
            username, url = self.auth()
        else:
            username, url = connection_info.username, connection_info.url
        arguments: schemas.ResponseData = self.connection.receive()
        self.parser = self.configure_parser(arguments.request_content)

        return username, url # return the username and url

    def interface(self, kwargs):
        if not kwargs['command_selection']:
            return
        command_request = schemas.CommandData(request_content=kwargs)
        self.connection.send(command_request)
        while True:
            command_response = self.connection.receive()
            if isinstance(command_response, schemas.ResponseData):
                self.url, self.username = command_response.url, command_response.active_username
                self.pwd, self.current_system = command_response.pwd, command_response.system
                if command_response.exit_status:
                    print("Exit initiated")
                    os._exit(0)
            handled_response, repeat = self.universal_message_handler(command_response, self.term)
            if not handled_response:
                break
            handled_response = schemas.FormResponse(request_content=handled_response)
            self.connection.send(handled_response)

    def terminal_cli(self):
        try:
            kwargs = self.parser.parse_args()
        except:
            print("Invalid Arguments")
            os._exit(0)
        kwargs = vars(kwargs)
        self.interface(kwargs)
        os._exit(0)

    def cli_window(self):
        title = pyfiglet.figlet_format("-----------\nTapisCLICICLE\n-----------", font="slant") # print the title when CLI is accessed
        print(title)
        print(r"""If you find any issues, please create a new issue here: https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/issues
                Enter 'exit' to exit the client
                Enter 'shutdown' to shut down the client and server
                Enter 'Help' to show command options""")
        while True:
            try:
                time.sleep(0.01)
                kwargs: dict = vars(self.parser.parse_args(str(input(f"[{self.username}@{self.url}][{self.current_system}]{self.pwd} ")).split(" ")))
                self.interface(kwargs)
            except KeyboardInterrupt:
                continue
            except (ConnectionAbortedError, ConnectionResetError):
                print("Server shutdown, exiting")
                os._exit(0)
            except (TypeError, argparse.ArgumentError, argparse.ArgumentTypeError) as e:
                print(e)
                error_str = traceback.format_exc()
                print(error_str)
                print("Invalid command")
            except Exception as e:
                error_str = traceback.format_exc()
                if self.debug:
                    print(error_str)
                error_message = schemas.ResponseData(error=str(e))
                self.connection.send(error_message)
        

    def main(self):
        if len(sys.argv) > 1: # checks if any command line arguments were provided. Does not open CLI
            self.terminal_cli()
        self.cli_window()


if __name__ == "__main__":
    client = CLI(socket.gethostbyname(socket.gethostname()), 30000)
    client.main()


import json


if __name__ != "__main__":
    from . import baseCommand, decorators


class create_app(baseCommand.BaseCommand):
    """
    @help: create an app. You must have a properly configured app config file. 
    See a template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/app-config.json
    @todo: add a system where the class can say whether or not the config file of the app exists, and if it does not, open up a form to get system information. Maybe 
    be able to run RequiresForm as something other than a decorator?
    """
    async def run(self, file: str, *args, **kwargs) -> str: # create a tapis app taking a json descriptor file path
        with open(file, 'r') as f:
            app_def = json.loads(f.read())
        url = self.t.apps.createAppVersion(**app_def)
        return f"App created successfully\nID: {app_def['id']}\nVersion: {app_def['version']}\nURL: {url}\n"


class get_apps(baseCommand.BaseCommand):
    """
    @help: get all the apps on a system
    """
    async def run(self, *args, **kwargs) -> str:
        apps = self.t.apps.getApps()
        return str(apps)


class delete_app(baseCommand.BaseCommand):
    """
    @help: delete the selected app
    """
    decorator=decorators.NeedsConfirmation()
    async def run(self, id: str, version: str, *args, **kwargs) -> str:
        return_value = self.t.apps.deleteApp(appId=id, appVersion=version)
        return str(return_value)


class get_app(baseCommand.BaseCommand):
    """
    @help: return selected app information
    """
    async def run(self, id: str, version: str, *args, **kwargs)-> None | str: # returns app information with an id and version as arguments
        app = self.t.apps.getApp(appId=id, appVersion=version)
        return app
    

class run_job(baseCommand.BaseCommand):
    """
    @help: run a job from an app on a system. You must have a properly configured job config file. 
    See a template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/job-config.json
    """
    async def run(self, file: str, *args, **kwargs)->str: # run a job using an app. Takes a job descriptor json file path
        with open(file, 'r') as f:
            app = json.loads(f.read())
        job = self.t.jobs.submitJob(**app)
        return str(job.uuid)


class get_job_status(baseCommand.BaseCommand):
    """
    @help: get the status of a job
    """
    async def run(self, uuid: str, *args, **kwargs)->str: # return a job status with its Uuid
        job_status = self.t.jobs.getJobStatus(jobUuid=uuid)
        return str(job_status)


class download_job_output(baseCommand.BaseCommand):
    """
    @help: download a job output from the system 
    """
    async def run(self, uuid: str, file: str, *args, **kwargs)->str: # download the output of a job with its Uuid
        jobs_output = self.t.jobs.getJobOutputDownload(jobUuid=uuid, outputPath='tapisjob.out')
        with open(file, 'w') as f:
            f.write(jobs_output)
        return f"Successfully downloaded job output to {file}"



from tapipy.tapis import Tapis


class ServiceChecker:
    def __supports_systems(self):
        if self.t.systems.getSchedulerProfiles():
            return True
        return False

    def check_services(self):
        return {
            "SYSTEM SUPPORT":self.__supports_systems(),
            "FILES SUPPORT":self.__supports_systems(),
            "APPS AND JOBS SUPPORT":self.__supports_systems(),
            "PODS SUPPORT":True
        }


import re
import inspect
import abc
import ast
import json
from typing import Type
import typing
from abc import abstractmethod, ABC

from commands import decorators # I finally understand. Imported at the top level by serverRun, so it can only see packages from that vantage point
from utilities import exceptions
from commands.arguments.argument import Argument, DynamicChoiceList, ALLOWED_ARG_TYPES
from socketopts import schemas


def get_kwargs(function):
    sig = inspect.signature(function)
    keyword_args = [param.name for param in sig.parameters.values() if param.default != inspect.Parameter.empty]
    return keyword_args

def get_args(function):
    sig = inspect.signature(function)
    positional_args = [param.name for param in sig.parameters.values() if param.default == inspect.Parameter.empty]
    return positional_args


class HelpStringRetriever:
    def help_string_retriever(self):
        try:
            docstring_components = self.__class__.__doc__
            docstring_components = docstring_components.split("@")
            for docstring_component in docstring_components:
                if re.match(r'^[^:]+', docstring_component).group(0) == "help":
                    return docstring_component.split("help: ")[1]
            raise Exception()
        except Exception as e:
            raise exceptions.HelpDoesNotExist(self.__class__.__name__)
    

class CommandMetaClass(abc.ABCMeta):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('BaseQuery', 'BaseCommand'):
            instance.__check_run(name, attrs)
            instance.__check_command_args(name, attrs)
            instance.__check_decorator(name, attrs)
            instance.__check_command_opt(name, attrs)
        return instance
    
    def __check_run(self, name, attrs):
        if 'run' not in list(attrs.keys()):
            raise AttributeError(f"The command {name} requires a 'run()' method")
        elif not inspect.iscoroutinefunction(attrs['run']):
            raise AttributeError(f"The run method of the command {name} must be a coroutine")
    
    def __check_command_args(self, name, attrs):
        run_params = get_args(attrs['run'])
        args = []
        if 'required_arguments' in list(attrs.keys()): 
            args += attrs['required_arguments']
        if 'optional_arguments' in list(attrs.keys()): 
            args += attrs['optional_arguments']
        if args:
            for argument in args:
                if not issubclass(argument.__class__, Argument) and not isinstance(argument, Argument):
                    raise AttributeError(f"The argument {argument.argument} of the command '{name}' must be of type 'Argument'")
                if 'optional_arguments' in list(attrs.keys()) and argument in attrs['optional_arguments'] and argument.positional:
                    raise AttributeError(f"The optional argument {argument.argument} of the command {name} cannot be positional. All positional arguments must be required")
        if 'kwargs' not in run_params or 'args' not in run_params:
            raise AttributeError(f"The run() method of the {name} class must have a **kwargs and *args as parameters")
        
    def __check_decorator(self, name, attrs):
        if 'decorator' in list(attrs.keys()) and type(attrs['decorator']) not in decorators.DECORATOR_LIST:
            raise TypeError(f"The decorator parameter of the command {name} is invalid. Must be set to None or to a decorator type. Currently is {type(attrs['decorator'])}")
        
    def __check_command_opt(self, name, attrs):
        if 'command_opt' in list(attrs.keys()) and type((attrs['command_opt'])) != list:
            raise TypeError(f"The command opt attribute of the command {name} must be a list!")


class HelpMenu:
    def __init__(self, required_arguments: dict[str, Argument], optional_arguments: dict[str, Argument]):
        self.required_arguments = required_arguments
        self.optional_arguments = optional_arguments
        self.arguments = {**required_arguments, **optional_arguments}
        self.help = self.create_help_menu()

    def create_help_menu(self):
        help_dict = {'required':dict(), 'optional':dict()}
        for arg_name, argument in self.required_arguments.items():
            help_dict['required'][arg_name] = argument.help_message()
        for arg_name, argument in self.optional_arguments.items():
            help_dict['optional'][arg_name] = argument.help_message()
        return help_dict
    
    def dict(self):
        return self.arguments


class BaseCommand(ABC, HelpStringRetriever, metaclass=CommandMetaClass):
    decorator = None
    return_formatter = None
    command_opt: list = None
    supports_config_file: bool = False
    required_arguments: list[Argument] | dict = list()
    optional_arguments: list[Argument] | dict = list()
    def __init__(self):
        self.t = None
        self.username = None
        self.password = None
        self.server = None
        self.arguments = dict()
        if isinstance(self.required_arguments, list):
            self.required_arguments = {argument.argument:argument for argument in self.required_arguments}
            self.required_arguments['connection'] = Argument('connection', arg_type='silent')
            self.arguments.update(**self.required_arguments)
        if isinstance(self.optional_arguments, list):  
            self.optional_arguments = {argument.argument:argument for argument in self.optional_arguments}
            self.arguments.update(**self.optional_arguments)
        self.positional_arguments = [arg_name for arg_name, arg in self.required_arguments.items() if arg.positional]
        self.form_arguments = [arg_name for arg_name, arg in self.arguments.items() if arg.arg_type not in  ('standard', 'silent')]
        self.help: dict[dict[str, list[dict[str, str]]]] = dict()

        self.command_execution_sequence = [self.verify_argument_rules_followed]
        if self.supports_config_file:
            self.command_execution_sequence.append(self.handle_config_file)
        if self.form_arguments:
            self.command_execution_sequence.append(self.handle_form_input)
        if self.positional_arguments:
            self.command_execution_sequence.append(self.check_for_positionals)
        if self.command_opt:
            self.command_execution_sequence.append(self.handle_arg_opts)
        self.command_execution_sequence.append(self.filter_kwargs)

    async def verify_argument_rules_followed(self, kwargs):
        for name, value in kwargs.items():
            if name in self.optional_arguments and value and name not in self.form_arguments:
                kwargs[name] = self.optional_arguments[name].verify_standard_value(value)
            elif name in self.required_arguments:
                kwargs[name] = self.required_arguments[name].verify_standard_value(value)
        return kwargs

    async def filter_kwargs(self, kwargs):
        filtered_kwargs = dict()
        for arg, value in kwargs.items():
            if arg in self.arguments and value != None:
                if (self.arguments[arg].arg_type in typing.get_args(ALLOWED_ARG_TYPES) and kwargs[arg]) or self.arguments[arg].arg_type == 'standard':
                    filtered_kwargs[arg] = value
        return filtered_kwargs

    async def handle_config_file(self, kwargs):
        if kwargs['file']:
            with open(kwargs['file'], 'r') as f:
                kwargs = json.loads(f.read)
        return kwargs
    
    async def handle_form_input(self, kwargs):
        for arg_name in self.form_arguments:
            if kwargs[arg_name] or arg_name in self.required_arguments:
                request = schemas.FormRequest(request_content={arg_name:self.arguments[arg_name] for arg_name in self.form_arguments if kwargs[arg_name] or arg_name in self.required_arguments})
                await kwargs['connection'].send(request)
                response: schemas.FormResponse = await kwargs['connection'].receive()
                kwargs.update(**response.request_content)
        print(kwargs)
        return kwargs
    
    async def handle_arg_opts(self, kwargs):
        for opt in self.command_opt:
            kwargs = opt(kwargs)
        return kwargs
    
    async def check_for_positionals(self, kwargs):
        for arg_name, value in zip(self.positional_arguments, kwargs['positionals']):
            kwargs[arg_name] = value
        return kwargs

    def set_t_and_creds(self, t, username, password, server):
        self.t = t
        self.username = username
        self.password = password
        self.server = server
        for key, arg in self.optional_arguments.items():
            if issubclass(arg.choices.__class__, DynamicChoiceList):
                arg.choices = arg.choices(self.t)

    def update_args_with_truncated(self, truncated_args_dict):
        try:
            for key in self.required_arguments:
                self.required_arguments[key].truncated_arg = f"-{truncated_args_dict[key]}"
            for key in self.optional_arguments:
                self.optional_arguments[key].truncated_arg = f"-{truncated_args_dict[key]}"
            self.help['required'] = self.__argument_help_compiler(self.required_arguments)
            self.help['optional'] = self.__argument_help_compiler(self.optional_arguments)
        except Exception as e:
            print(self.__class__.__name__)
            raise e
        
    def __argument_help_compiler(self, arg_dict: dict[str, Argument]):
        verbose_dict = dict()
        standard_str = str()
        for name, argument in arg_dict.items():
            if not (name in self.required_arguments and name in self.form_arguments) and argument.arg_type != 'silent':
                verbose_dict[name] = argument.help_message()
                standard_str += argument.str()
        return {'verbose':verbose_dict, 'standard':standard_str}

    def __get_help(self, verbose):
        help_dict = {"Command":self.__class__.__name__,
                    "Description":self.help_string_retriever()}
        if not verbose:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__} {self.help['required']['standard']}\n(Optional Arguments) {self.help['optional']['standard']}"})
        else:
            help_dict.update(**{
                "Syntax":f"{self.__class__.__name__}",
                "Required Arguments":self.help['required']['verbose'],
                "Optional Arguments":self.help['optional']['verbose']})
        return help_dict
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        pass

    async def __call__(self, **kwargs):
        command = kwargs.pop('command_selection')

        if self.decorator:
            return_value = await self.decorator(input_command=self, **kwargs)
        
        verbose = kwargs.pop('verbose')
        help = kwargs.pop('help')

        if help:
            return self.__get_help(verbose=verbose)
        for handler in self.command_execution_sequence:
            kwargs = await handler(kwargs)
        else:
            return_value = await self.run(**kwargs)
        if self.return_formatter:
            return_value = self.return_formatter(return_value, verbose)
        return return_value
    

class BaseQuery(BaseCommand):
    def get_pod_credentials(self, id):
        uname, pword = self.t.pods.get_pod_credentials(pod_id=id).user_username, self.t.pods.get_pod_credentials(pod_id=id).user_password
        return uname, pword
    
    @abstractmethod
    async def run(self, *args, **kwargs):
        pass
    

class CommandMapMetaClass(type):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        if name not in ('CommandMapMetaClass', 'BaseCommandMap'):
            instance.__check_commands_are_proper_type(name, attrs)
            instance.__check_command_name(name, attrs)
            instance.__check_data_formatter(name, attrs)
            instance.__special_command_opts(name, attrs)
        return instance
    
    def __check_commands_are_proper_type(self, name, attrs):
        commands = attrs['command_map']
        if not commands:
            raise AttributeError(f"The command group {name} has no commands!")
        for command_name, command in commands.items():
            if not issubclass(command.__class__, BaseCommand):
                raise AttributeError(f"The command {command_name} was not passed to the {name} as a class, but as an object. Ensure you do not instantiate commands when defining the class")
            
    def __check_command_name(self, name, attrs):
        commands = attrs['command_map']
        for command_name, command in commands.items():
            if command_name != command.__class__.__name__:
                raise AttributeError(f"The command {command_name} in the command map {name} is a different name from its corresponding command class, {command.__class__.__name__}")
            
    def __check_data_formatter(self, name, attrs):
        if 'data_formatter' not in list(attrs.keys()):
            print(f"WARNING: The command map {name} has no data formatter. If any commands have non json-serializable return values, command execution will fail! These must be handled by a formatter")

    def __special_command_opts(self, name, attrs):
        if 'command_opt' in list(attrs.keys()):
            if type(attrs['command_opt']) != list:
                raise TypeError(f"The command_opt attribute for the command group {name} must be a list!")
            for command_name, command in attrs['command_map'].items():
                if not command.command_opt:
                    command.command_opt = attrs['command_opt']
                    continue
                command.command_opt += attrs['command_opt']


class CommandContainer:
    aggregate_command_map: dict[str, Type[BaseCommand]] = dict()
    arguments : dict[str, Argument]= {
        'help':Argument('help', action='store_true'),
        'verbose':Argument('verbose', action='store_true'),
        'file':Argument('file')
    }


class BaseCommandMap(CommandContainer, HelpStringRetriever, metaclass=CommandMapMetaClass):
    command_map: dict[str, Type[BaseCommand]] = None
    data_formatter = None
    command_opt = None
    def __init__(self):
        self.brief_help = self.__brief_help_gen()
        self.verbose_help = self.__help_gen()
        for name, command in self.command_map.items():
            command.return_formatter = self.data_formatter
            self.aggregate_command_map.update({name:command})
            if command.required_arguments or command.optional_arguments:
                self.__contribute_to_arguments(command)

    def __contribute_to_arguments(self, command):
        arg_dict = {**command.required_arguments, **command.optional_arguments}
        for name, arg in arg_dict.items():
            if name not in self.arguments:
                self.arguments[name] = arg
            elif self.arguments[name].check_for_copy_data() != arg.check_for_copy_data():
                raise ValueError(f"Found two instances of the argument {arg.argument} that had different attributes.\n\n{command.__class__.__name__}: {arg.check_for_copy_data()}\n\nvs\n\n{self.arguments[name].check_for_copy_data()}")

    def __help_gen(self):
        verbose_help = dict()
        for command in self.command_map.values():
            verbose_help.update({command.__class__.__name__:command.help})
        return verbose_help
    
    def __brief_help_gen(self):
        brief_help = {
            'group':self.__class__.__name__,
            'description':self.help_string_retriever(),
            'instruction':f"enter {self.__class__.__name__} to retrieve list of commands"
        }
        return brief_help
    
    def __call__(self):
        return self.verbose_help



import sys
import time
import socket
import os
import asyncio
import traceback

from tapipy.tapis import Tapis

if __name__ != "__main__":
    from commands import commandMap, decorators
    from utilities import logger, exceptions
    from socketopts import schemas, socketOpts
    from server import auth
else:
    import commands.commandMap as commandMap


class ServerConnection(socketOpts.ServerSocketOpts):
    """
    connection object to wrap around async reader and writer to make work easier
    """
    def __init__(self, name, reader, writer, debug=False):
        super().__init__(name, debug=debug)
        self.name = name
        self.reader = reader
        self.writer = writer

    async def close(self):
        self.writer.close()
        await self.writer.wait_closed()
        

class TaskCallback:
    def __init__(self, logger, task: asyncio.Task):
        self.logger, self.task = logger, task
        print("DOING THE INIT")

    def __call__(self, result):
        print(result)
        result = self.task.result()
        self.logger.info(result)


class Server(commandMap.AggregateCommandMap, logger.ServerLogger, decorators.DecoratorSetup, auth.ServerSideAuth):
    """
    Receives commands from the client and executes Tapis operations
    """
    SESSION_TIME = 1200
    debug=False
    def __init__(self, IP: str, PORT: int):
        super().__init__()
        self.initial = True

        self.t = None
        self.url = None
        self.access_token = None
        self.username = None
        self.password = None
        self.auth_type = None

        self.__name__ = "Server"
        self.initialize_logger(self.__name__)
        # setting up socket server
        self.ip, self.port = IP, PORT
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(False)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.ip, self.port))
        self.sock.listen(1)
        self.end_time = time.time() + self.SESSION_TIME # start the countdown on the timeout

        self.task_list: list[asyncio.Task] = []

        self.server = None
        self.num_connections = 0

        self.logger.info('initialization complete')

    def cancel_tasks(self):
        for task in self.task_list:
            task.cancel()

    async def handshake(self, connection):
        self.logger.info("Handshake starting")
        if self.initial:  # if this is the first time in the session that the cli is connecting
            startup_data = schemas.StartupData(initial = self.initial)
            await connection.send(startup_data)
            await self.auth_startup(connection)
        else:
            startup_result = schemas.StartupData(initial = self.initial, username = self.username, url = self.url)
            await connection.send(startup_result)
        self.initial = False
        self.logger.info("Final connection data sent")

    async def accept(self, reader, writer):
        """
        accept connection request and initialize communication with the client
        """  
        self.num_connections += 1
        connection = ServerConnection(f"CON-{self.num_connections}", reader=reader, writer=writer, debug=self.debug)
        self.timeout_handler()
        ip, port= writer.transport.get_extra_info('socket').getsockname()
        if ip != socket.gethostbyname(socket.gethostname()):
            raise exceptions.UnauthorizedAccessError(ip)
        self.logger.info("Received connection request")
        try:
            await self.handshake(connection)
        except exceptions.InvalidCredentialsReceived as e:
            self.logger.warning("invalid credentials entered too many times. Cancelling request")
            await connection.close()
            return
        except exceptions.ClientSideError as e:
            self.logger.warning(f"Encountered client side error during startup handshake. {e}")
            await connection.close()
            return
        except Exception as e:
            self.logger.warning(e)
            await connection.close()
            return
        self.logger.info("connection is running now")
        await connection.send(schemas.ResponseData(request_content={name:argument.json() for name, argument in self.arguments.items()}))
        
        setup_response: schemas.ResponseData = await connection.receive()
        if not setup_response.request_content['setup_success']:
            self.logger.warning(f"The setup of the connection {connection.name} failed")
            return

        loop = asyncio.get_event_loop()
        task: asyncio.Task = loop.create_task(self.receive_and_execute(connection))
        callback = TaskCallback(self.logger, task)
        task.add_done_callback(callback)
        self.task_list.append(task)

    def timeout_handler(self):  
        """
        checks if the timeout has been exceeded
        """
        if time.time() > self.end_time: 
            raise exceptions.TimeoutError

    async def receive_and_execute(self, connection):
        """
        receive and process commands
        """
        while True:
            try:
                message = await connection.receive()
                self.timeout_handler()  
                kwargs = message.request_content
                result = await self.run_command(connection, kwargs)
                response = schemas.ResponseData(message={"message":result}, url=self.url, active_username=self.username)
                self.end_time = time.time() + self.SESSION_TIME 
                await connection.send(response)
                self.logger.info(message.schema_type)
            except exceptions.ClientSideError as e:
                self.logger.warning(e)
                continue
            except (exceptions.TimeoutError, exceptions.Shutdown) as e:
                error_response = schemas.ResponseData(error=str(e), exit_status=1, url=self.url, active_username=self.username)
                await connection.send(error_response)
                self.logger.warning(str(e))
                self.cancel_tasks()
                self.server.close()
                loop = asyncio.get_event_loop()
                loop.stop()
                loop.close()
                sys.exit(0)
            except exceptions.Exit as e:
                self.logger.info("user exit initiated")
                error_response = schemas.ResponseData(error=str(e), exit_status=1, url=self.url, active_username=self.username)
                await connection.send(error_response)
                await connection.close()
                return
                #self.accept()  # wait for CLI to reconnect
            except OSError:
                self.logger.info("connection was lost, waiting to reconnect")
                await connection.close()
            except (exceptions.CommandNotFoundError, exceptions.NoConfirmationError, exceptions.InvalidCredentialsReceived, Exception) as e:
                error_str = traceback.format_exc()
                error_response = schemas.ResponseData(error=str(e), url=self.url, active_username=self.username)
                await connection.send(error_response)
                self.logger.warning(f"{error_str}")
    
    async def main(self):
        self.server = await asyncio.start_server(self.accept, sock=self.sock)
        try:
            async with self.server:
                await self.server.serve_forever()
        except KeyboardInterrupt:
            self.server.close()
            self.cancel_tasks()
            sys.exit(0)


if __name__ == '__main__':
    server = Server(socket.gethostbyname(socket.gethostname()), 30000)
    asyncio.run(server.main())
